import unittest
from .testHelpers import get_cpp_fileinfo, get_cpp_function_list

class TestCyclomaticComplexity(unittest.TestCase):

    def test_one_function_with_no_condition(self):
        result = get_cpp_function_list("int fun(){}")
        self.assertEqual(1, result[0].cyclomatic_complexity)

    def test_one_function_with_one_condition(self):
        result = get_cpp_function_list("int fun(){if(a){xx;}}")
        self.assertEqual(2, result[0].cyclomatic_complexity)

    def test_one_function_with_question_mark(self):
        result = get_cpp_function_list("int fun(){return (a)?b:c;}")
        self.assertEqual(2, result[0].cyclomatic_complexity)

    def test_one_function_with_forever_loop(self):
        result = get_cpp_function_list("int fun(){for(;;){dosomething();}}")
        self.assertEqual(2, result[0].cyclomatic_complexity)

    def test_one_function_with_and(self):
        result = get_cpp_function_list("int fun(){if(a&&b){xx;}}")
        self.assertEqual(3, result[0].cyclomatic_complexity)

    def test_one_function_with_else_if(self):
        result = get_cpp_function_list("int fun(){if(a)b;else if (c) d;}")
        self.assertEqual(3, result[0].cyclomatic_complexity)

    def test_sharp_if_and_sharp_elif_counts_in_cc_number(self):
        result = get_cpp_function_list('''
                int main(){
                #ifdef A
                #elif (defined E)
                #endif
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual(3, result[0].cyclomatic_complexity)




import unittest
import inspect
from lizard import analyze_file, FileAnalyzer, get_extensions


def get_swift_function_list(source_code):
    return analyze_file.analyze_source_code(
        "a.swift", source_code).function_list


class Test_parser_for_Swift(unittest.TestCase):

    def test_empty(self):
        functions = get_swift_function_list("")
        self.assertEqual(0, len(functions))

    def test_no_function(self):
        result = get_swift_function_list('''
            for name in names {
                print("Hello, \(name)!")
            }
                ''')
        self.assertEqual(0, len(result))

    def test_one_function(self):
        result = get_swift_function_list('''
            func sayGoodbye() { }
                ''')
        self.assertEqual(1, len(result))
        self.assertEqual("sayGoodbye", result[0].name)
        self.assertEqual(0, result[0].parameter_count)
        self.assertEqual(1, result[0].cyclomatic_complexity)

    def test_one_with_parameter(self):
        result = get_swift_function_list('''
            func sayGoodbye(personName: String, alreadyGreeted: Bool) { }
                ''')
        self.assertEqual(1, len(result))
        self.assertEqual("sayGoodbye", result[0].name)
        self.assertEqual(2, result[0].parameter_count)

    def test_one_function_with_return_value(self):
        result = get_swift_function_list('''
            func sayGoodbye() -> String { }
                ''')
        self.assertEqual(1, len(result))
        self.assertEqual("sayGoodbye", result[0].name)

    def test_one_function_with_complexity(self):
        result = get_swift_function_list('''
            func sayGoodbye() { if ++diceRoll == 7 { diceRoll = 1 }}
                ''')
        self.assertEqual(2, result[0].cyclomatic_complexity)

    def test_interface(self):
        result = get_swift_function_list('''
            protocol p {
                func f1() -> Double
                func f2() -> NSDate
            }
            func sayGoodbye() { }
                ''')
        self.assertEqual(1, len(result))
        self.assertEqual("sayGoodbye", result[0].name)

    def test_interface_followed_by_a_class(self):
        result = get_swift_function_list('''
            protocol p {
                func f1() -> Double
                func f2() -> NSDate
            }
            class c { }
                ''')
        self.assertEqual(0, len(result))

    def test_interface_with_var(self):
        result = get_swift_function_list('''
            protocol p {
                func f1() -> Double
                var area: Double { get }
            }
            class c { }
                ''')
        self.assertEqual(0, len(result))

    def test_interface_with_var(self):
        result = get_swift_function_list('''
            protocol p {
                func f1() -> Double
                var area: Double { get }
            }
            class c { }
                ''')
        self.assertEqual(0, len(result))

    def test_generic_function(self):
        result = get_swift_function_list('''
            func f<T>() {}
                ''')
        self.assertEqual("f", result[0].name)

    def xtest_generic_function(self):
        result = get_swift_function_list('''
        func f<C1, C2: Container where (C1.t == C2.t)> (c1: C1, c: C2) -> Bool {}
                ''')
        self.assertEqual("f", result[0].name)
        self.assertEqual(2, result[0].parameter_count)




import unittest
from lizard import CLikeReader, CLikeReader
from mock import Mock
from .testHelpers import get_cpp_fileinfo, get_cpp_function_list

class Test_C_Token_extension(unittest.TestCase):

    def test_connecting_marcro(self):
        extended = CLikeReader(None).preprocess(("a##b c", ))
        #tbd

class Test_c_cpp_lizard(unittest.TestCase):

    def test_empty(self):
        result = get_cpp_function_list("")
        self.assertEqual(0, len(result))

    def test_no_function(self):
        result = get_cpp_function_list("#include <stdio.h>\n")
        self.assertEqual(0, len(result))

    def test_one_function(self):
        result = get_cpp_function_list("int fun(){}")
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)

    def test_two_function(self):
        result = get_cpp_function_list("int fun(){}\nint fun1(){}\n")
        self.assertEqual(2, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun1", result[1].name)
        self.assertEqual(1, result[0].start_line)
        self.assertEqual(1, result[0].end_line)
        self.assertEqual(2, result[1].start_line)
        self.assertEqual(2, result[1].end_line)

    def test_function_with_content(self):
        result = get_cpp_function_list("int fun(xx oo){int a; a= call(p1,p2);}")
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( xx oo )", result[0].long_name)

    def test_old_style_c_function(self):
        result = get_cpp_function_list("""int fun(param) int praram; {}""")
        self.assertEqual(1, len(result))

    def test_not_old_style_c_function(self):
        result = get_cpp_function_list("m()"*1500+ "a(){}")
        self.assertEqual(1, len(result))

    def test_complicated_c_function(self):
        result = get_cpp_function_list("""int f(int(*)()){}""")
        self.assertEqual('f', result[0].name)

    def test_function_dec_with_throw(self):
        result = get_cpp_function_list("""int fun() throw();void foo(){}""")
        self.assertEqual(1, len(result))

    def test_function_dec_with_noexcept(self):
        result = get_cpp_function_list("int fun() noexcept(true);void foo(){}")
        self.assertEqual(1, len(result))

    def test_function_dec_followed_with_one_word_is_ok(self):
        result = get_cpp_function_list("""int fun() no_throw {}""")
        self.assertEqual(1, len(result))

    def test_function_declaration_is_not_counted(self):
        result = get_cpp_function_list("""int fun();class A{};""")
        self.assertEqual(0, len(result))

    def test_old_style_c_function_has_semicolon(self):
        result = get_cpp_function_list("""{(void*)a}{}""")
        self.assertEqual(0, len(result))

    def test_typedef_is_not_old_style_c_function(self):
        result = get_cpp_function_list('''typedef T() nT; foo(){}''')
        self.assertEqual("foo", result[0].name)

    def test_stupid_macro_before_function(self):
        result = get_cpp_function_list('''T() foo(){}''')
        self.assertEqual("foo", result[0].name)

    def test_only_word_can_be_function_name(self):
        result = get_cpp_function_list("""[(){}""")
        self.assertEqual(0, len(result))

    def test_double_slash_within_string(self):
        result = get_cpp_function_list("""int fun(){char *a="\\\\";}""")
        self.assertEqual(1, len(result))

    def test_function_with_no_param(self):
        result = get_cpp_function_list("int fun(){}")
        self.assertEqual(0, result[0].parameter_count)

    def test_function_with_1_param(self):
        result = get_cpp_function_list("int fun(aa * bb){}")
        self.assertEqual(1, result[0].parameter_count)
        self.assertEqual(["bb"], result[0].parameters)

    def test_function_with_param(self):
        result = get_cpp_function_list("int fun(aa * bb, cc dd){}")
        self.assertEqual(2, result[0].parameter_count)

    def test_function_with_strang_param(self):
        result = get_cpp_function_list("int fun(aa<mm, nn> bb){}")
        self.assertEqual(1, result[0].parameter_count)

    def test_one_function_with_namespace(self):
        result = get_cpp_function_list("int abc::fun(){}")
        self.assertEqual(1, len(result))
        self.assertEqual("abc::fun", result[0].name)
        self.assertEqual("abc::fun( )", result[0].long_name)

    def test_one_function_with_const(self):
        result = get_cpp_function_list("int abc::fun()const{}")
        self.assertEqual(1, len(result))
        self.assertEqual("abc::fun", result[0].name)
        self.assertEqual("abc::fun( ) const", result[0].long_name)

    def test_one_function_with_noexcept(self):
        result = get_cpp_function_list("int abc::fun()noexcept{}")
        self.assertEqual(1, len(result))
        self.assertEqual("abc::fun", result[0].name)
        self.assertEqual("abc::fun( ) noexcept", result[0].long_name)

    def test_one_function_in_class(self):
        result = get_cpp_function_list("class c {~c(){}}; int d(){}")
        self.assertEqual(2, len(result))
        self.assertEqual("c::~c", result[0].name)
        self.assertEqual("d", result[1].name)

    def test_one_macro_in_class(self):
        result = get_cpp_function_list("class c {M()}; int d(){}")
        self.assertEqual(1, len(result))
        self.assertEqual("d", result[0].name)

    def test_pre_class(self):
        result = get_cpp_function_list("class c; int d(){}")
        self.assertEqual(1, len(result))
        self.assertEqual("d", result[0].name)

    def test_class_with_inheritance(self):
        result = get_cpp_function_list("class c final:public b {int f(){}};")
        self.assertEqual(1, len(result))
        self.assertEqual("c::f", result[0].name)

    def test_nested_class(self):
        result = get_cpp_function_list("class c {class d {int f(){}};};")
        self.assertEqual(1, len(result))
        self.assertEqual("c::d::f", result[0].name)

    def test_braket_that_is_not_a_namespace(self):
        result = get_cpp_function_list("class c { {};int f(){}};")
        self.assertEqual(1, len(result))
        self.assertEqual("c::f", result[0].name)

    def test_nested_class_middle(self):
        result = get_cpp_function_list("class c {class d {};int f(){}};")
        self.assertEqual(1, len(result))
        self.assertEqual("c::f", result[0].name)

    def test_template_as_reference(self):
        result = get_cpp_function_list("abc::def(a<b>& c){}")
        self.assertEqual(1, len(result))

    def test_less_then_is_not_template(self):
        result = get_cpp_function_list("def(<); foo(){}")
        self.assertEqual(1, len(result))

    def test_template_with_pointer(self):
        result = get_cpp_function_list("abc::def (a<b*> c){}")
        self.assertEqual(1, len(result))

    def test_nested_template(self):
        result = get_cpp_function_list("abc::def (a<b<c>> c){}")
        self.assertEqual(1, len(result))

    def test_double_nested_template(self):
        result = get_cpp_function_list("abc::def (a<b<c<d>>> c){}")
        self.assertEqual(1, len(result))

    def test_template_with_reference(self):
        result = get_cpp_function_list("void fun(t<int &>b){} ")
        self.assertEqual(1, len(result))

    def test_template_with_reference_as_reference(self):
        result = get_cpp_function_list("void fun(t<const int&>&b){} ")
        self.assertEqual(1, len(result))

    def test_template_as_part_of_function_name(self):
        result = get_cpp_function_list("void fun<a,b<c>>(){} ")
        self.assertEqual('fun<a,b<c>>', result[0].name)

    def test_operator_overloading(self):
        result = get_cpp_function_list("bool operator +=(int b){}")
        self.assertEqual("operator +=", result[0].name)

    def test_operator_with_complicated_name(self):
        result = get_cpp_function_list("operator MyStruct&(){}")
        self.assertEqual("operator MyStruct &", result[0].name)

    def test_operator_overloading_with_namespace(self):
        result = get_cpp_function_list("bool TC::operator !(int b){}")
        self.assertEqual(1, len(result))
        self.assertEqual("TC::operator !", result[0].name)

    def test_function_operator(self):
        result = get_cpp_function_list("bool TC::operator ()(int b){}")
        self.assertEqual(1, len(result))
        self.assertEqual("TC::operator ( )", result[0].name)

    def test_inline_operator(self):
        result = get_cpp_function_list("class A { bool operator()(int b) {} };")
        self.assertEqual(1, len(result))
        self.assertEqual("A::operator ( )", result[0].name)

    def test_namespace_alias(self):
        result = get_cpp_function_list(
                "namespace p;"
                "namespace real { bool foo() {} }")
        self.assertEqual(1, len(result))
        self.assertEqual("real::foo", result[0].name)

    def test_nested_unnamed_namespace(self):
        result = get_cpp_function_list(
                "namespace real { namespace { bool foo() {} } }")
        self.assertEqual(1, len(result))
        self.assertEqual("real::foo", result[0].name)

    def test_constructor_initialization_list(self):
        result = get_cpp_function_list('''A::A():a(1){}''')
        self.assertEqual(1, len(result))
        self.assertEqual("A::A", result[0].name)

    def test_constructor_initialization_list_noexcept(self):
        result = get_cpp_function_list('''A::A()noexcept:a(1){}''')

    def test_constructor_initializer_list(self):
        result = get_cpp_function_list('''A::A():a({1}),value(true){}''')
        self.assertEqual(1, len(result))
        self.assertEqual("A::A", result[0].name)

    def test_constructor_uniform_initialization(self):
        result = get_cpp_function_list('''A::A():a{1}{}''')
        self.assertEqual(1, len(result))
        self.assertEqual("A::A", result[0].name)

    def test_brakets_before_function(self):
        result = get_cpp_function_list('''()''')
        self.assertEqual(0, len(result))

    def test_destructor_implementation(self):
        result = get_cpp_function_list('''A::~A(){}''')
        self.assertEqual(1, len(result))
        self.assertEqual("A::~A", result[0].name)

    def test_function_that_returns_function_pointers(self):
        result = get_cpp_function_list('''int (*fun())(){}''')
        self.assertEqual(1, len(result))
        self.assertEqual("int( * fun ( ) )", result[0].name)

    def test_underscore(self):
        from lizard import CodeReader
        generate_tokens = CodeReader.generate_tokens
        result = get_cpp_function_list(''' a() _() { }''')
        self.assertEqual(1, len(result))

class Test_Preprocessing(unittest.TestCase):

    def test_content_macro_should_be_ignored(self):
        result = get_cpp_function_list(r'''
                    #define MTP_CHEC                    \
                       int foo () {                     \
                        }
               ''')
        self.assertEqual(0, len(result))


    def test_preprocessors_should_be_ignored_outside_function_implementation(self):
        result = get_cpp_function_list('''
                      #ifdef MAGIC
                      #endif
                      void foo()
                      {}
                    ''')
        self.assertEqual(1, len(result))

    def test_preprocessor_is_not_function(self):
        result = get_cpp_function_list('''
                #ifdef A
                #elif (defined E)
                #endif
                ''')
        self.assertEqual(0, len(result))


class Test_Big(unittest.TestCase):

    def test_trouble(self):
        code = "foo<y () >> 5> r;"
        result = get_cpp_function_list(code)
        self.assertEqual(0, len(result))

    def test_namespace_is_not_function(self):
        code = "namespace a b(){}"
        result = get_cpp_function_list(code)
        self.assertEqual(0, len(result))



import unittest
from .testHelpers import get_cpp_function_list_with_extnesion
from lizard_ext.lizarddependencycount import LizardExtension as DependencyCounter


class TestFunctionDependencyCount(unittest.TestCase):

    def test_no_return(self):
        result = get_cpp_function_list_with_extnesion(
            "int fun(){}",
            DependencyCounter())
        self.assertEqual(0, result[0].dependency_count)

    def test_import_dependency(self):
        result = get_cpp_function_list_with_extnesion(
            "import library; int fun(){library.callMethod()}",
            DependencyCounter())
        self.assertEqual(1, result[0].dependency_count)

    def test_python_import_as(self):
        result = get_cpp_function_list_with_extnesion(
            "import python as py; int fun(){py.callMethod() py.version = 99}",
            DependencyCounter())
        self.assertEqual(2, result[0].dependency_count)
        result = get_cpp_function_list_with_extnesion(
            "import candy int fun(){candy.callMethod() py.version = 99}",
            DependencyCounter())
        self.assertEqual(1, result[0].dependency_count)
        result = get_cpp_function_list_with_extnesion(
            "import kok as www import tree, java as monster, coffee import java public board { private void function() { java += 1; java.tree = green; www.yay(0); teacher.lecture(0); }",
            DependencyCounter())
        self.assertEqual(3, result[0].dependency_count)



'''
Language parser for JavaScript
'''

from lizard import CLikeReader, CodeReader


class JavaReader(CLikeReader, CodeReader):
    # pylint: disable=R0903

    ext = ['java']
    language_names = ['java']

    def __init__(self, context):
        super(JavaReader, self).__init__(context)

    def _state_old_c_params(self, token):
        if token == '{':
            self._state_dec_to_imp(token)

    def _state_global(self, token):
        if token == '@':
            self._state = self._state_decorator
            return
        super(JavaReader, self)._state_global(token)

    def _state_decorator(self, _):
        self._state = self._state_post_decorator

    def _state_post_decorator(self, token):
        if token == '.':
            self._state = self._state_decorator
        else:
            self._state = self._state_global
            self._state(token)



'''
Thanks for Holy Wen from Nokia Siemens Networks to let me use his code
to put the result into xml file that is compatible with cppncss.
Jenkens has plugin for cppncss format result to display the diagram.
'''


def xml_output(result, verbose):
    import xml.dom.minidom

    impl = xml.dom.minidom.getDOMImplementation()
    doc = impl.createDocument(None, "cppncss", None)
    root = doc.documentElement

    processing_instruction = doc.createProcessingInstruction(
        'xml-stylesheet',
        'type="text/xsl" ' +
        'href="https://raw.github.com/terryyin/lizard/master/lizard.xsl"')
    doc.insertBefore(processing_instruction, root)

    root.appendChild(_create_function_measure(doc, result, verbose))
    root.appendChild(_create_file_measure(doc, result))

    return doc.toprettyxml()


def _create_function_measure(doc, result, verbose):
    measure = doc.createElement("measure")
    measure.setAttribute("type", "Function")
    measure.appendChild(_create_labels(doc, ["Nr.", "NCSS", "CCN"]))

    number = 0
    total_func_ncss = 0
    total_func_ccn = 0

    for source_file in result:
        file_name = source_file.filename
        for func in source_file.function_list:
            number += 1
            total_func_ncss += func.nloc
            total_func_ccn += func.cyclomatic_complexity
            measure.appendChild(
                _create_function_item(
                    doc, number, file_name, func, verbose))

        if number != 0:
            measure.appendChild(
                _create_labeled_value_item(
                    doc, 'average', "NCSS", str(total_func_ncss / number)))
            measure.appendChild(
                _create_labeled_value_item(
                    doc, 'average', "CCN", str(total_func_ccn / number)))
    return measure


def _create_file_measure(doc, result):
    measure = doc.createElement("measure")
    measure.setAttribute("type", "File")
    measure.appendChild(
        _create_labels(doc, ["Nr.", "NCSS", "CCN", "Functions"]))

    file_nr = 0
    file_total_ncss = 0
    file_total_ccn = 0
    file_total_funcs = 0

    for source_file in result:
        file_nr += 1
        file_total_ncss += source_file.nloc
        file_total_ccn += source_file.CCN
        file_total_funcs += len(source_file.function_list)
        measure.appendChild(
            _create_file_node(doc, source_file, file_nr))

    if file_nr != 0:
        file_summary = [("NCSS", file_total_ncss / file_nr),
                        ("CCN", file_total_ccn / file_nr),
                        ("Functions", file_total_funcs / file_nr)]
        for key, val in file_summary:
            measure.appendChild(
                _create_labeled_value_item(doc, 'average', key, val))

    summary = [("NCSS", file_total_ncss),
               ("CCN", file_total_ccn),
               ("Functions", file_total_funcs)]
    for key, val in summary:
        measure.appendChild(_create_labeled_value_item(doc, 'sum', key, val))
    return measure


def _create_label(doc, name):
    label = doc.createElement("label")
    text1 = doc.createTextNode(name)
    label.appendChild(text1)
    return label


def _create_labels(doc, label_name):
    labels = doc.createElement("labels")
    for label in label_name:
        labels.appendChild(_create_label(doc, label))

    return labels


def _create_function_item(doc, number, file_name, func, verbose):
    item = doc.createElement("item")
    if verbose:
        item.setAttribute(
            "name", "%s at %s:%s" %
            (func.long_name, file_name, func.start_line))
    else:
        item.setAttribute(
            "name", "%s(...) at %s:%s" %
            (func.name, file_name, func.start_line))
    value1 = doc.createElement("value")
    text1 = doc.createTextNode(str(number))
    value1.appendChild(text1)
    item.appendChild(value1)
    value2 = doc.createElement("value")
    text2 = doc.createTextNode(str(func.nloc))
    value2.appendChild(text2)
    item.appendChild(value2)
    value3 = doc.createElement("value")
    text3 = doc.createTextNode(str(func.cyclomatic_complexity))
    value3.appendChild(text3)
    item.appendChild(value3)
    return item


def _create_labeled_value_item(doc, name, label, value):
    average_ncss = doc.createElement(name)
    average_ncss.setAttribute("lable", label)
    average_ncss.setAttribute("value", str(value))
    return average_ncss


def _create_file_node(doc, source_file, file_nr):
    item = doc.createElement("item")
    item.setAttribute("name", source_file.filename)
    value1 = doc.createElement("value")
    text1 = doc.createTextNode(str(file_nr))
    value1.appendChild(text1)
    item.appendChild(value1)
    value2 = doc.createElement("value")
    text2 = doc.createTextNode(str(source_file.nloc))
    value2.appendChild(text2)
    item.appendChild(value2)
    value3 = doc.createElement("value")
    text3 = doc.createTextNode(str(source_file.CCN))
    value3.appendChild(text3)
    item.appendChild(value3)
    value4 = doc.createElement("value")
    text4 = doc.createTextNode(str(len(source_file.function_list)))
    value4.appendChild(text4)
    item.appendChild(value4)
    return item



''' Language parser for Python '''

from lizard import CodeReader
from lizard import CLikeReader


class ObjCReader(CLikeReader, CodeReader):

    ext = ['m']
    language_names = ['objectivec', 'objective-c', 'objc']

    def __init__(self, context):
        super(ObjCReader, self).__init__(context)

    def fake_and_useless(self):
        pass

    def useless_and_fake(self):
        pass

    def _state_global(self, token):
        super(ObjCReader, self)._state_global(token)
        if token == '(':
            self.bracket_stack.append(token)
            self._state = self._state_dec
            self.context.add_to_long_function_name(token)

    def _state_dec_to_imp(self, token):
        if token in ("+", "-"):
            self._state = self._state_global
        else:
            super(ObjCReader, self)._state_dec_to_imp(token)
            if self._state != self._state_imp:
                self._state = self._state_objc_dec_begin
                self.context.start_new_function(token)

    def _state_objc_dec_begin(self, token):
        if token == ':':
            self._state = self._state_objc_dec
            self.context.add_to_function_name(token)
        elif token == '{':
            self.br_count += 1
            self._state = self._state_imp
        else:
            self._state = self._state_global

    def _state_objc_dec(self, token):
        if token == '(':
            self._state = self._state_objc_param_type
            self.context.add_to_long_function_name(token)
        elif token == ',':
            pass
        elif token == '{':
            self.br_count += 1
            self._state = self._state_imp
        else:
            self._state = self._state_objc_dec_begin
            self.context.add_to_function_name(" " + token)

    def _state_objc_param_type(self, token):
        if token == ')':
            self._state = self._state_objc_param
        self.context.add_to_long_function_name(" " + token)

    def _state_objc_param(self, _):
        self._state = self._state_objc_dec



''' Language parser for TTCN-3 '''

import re
from lizard import CLikeReader, CodeReader


class TTCNReader(CLikeReader, CodeReader):  # pylint: disable=R0903

    ext = ['ttcn', 'ttcnpp']
    language_names = ['ttcn', 'ttcn3']

    conditions = set(['if', 'else', 'for', 'while', 'altstep',
                      'case', 'goto', 'alt', 'interleave',
                      'and', 'or', 'xor'])

    parameter_bracket_open = '('
    parameter_bracket_close = ')'

    def __init__(self, context):
        super(TTCNReader, self).__init__(context)

    # module and group blocks are ignored
    def _state_global(self, token):
        if token == 'testcase':
            self._state = self._state_function
            self.context.start_new_function('__testcase__')
        elif token == 'function':
            self._state = self._state_function
            self.context.start_new_function('')
        elif token == 'control':
            self.context.start_new_function('__control__')
            self._state = self._state_dec_to_imp

    def _state_function(self, token):
        if token[0].isalpha():
            self.context.add_to_function_name(token)
        elif token == '(':
            self.bracket_stack.append(token)
            self._state = self._state_dec
            self.context.add_to_long_function_name(token)
        elif token == '@deterministic':
            self.context.add_to_long_function_name(token + ' ')
        else:
            self._state = self._state_global

    def _state_dec_to_imp(self, token):
        if token == '{':
            self.br_count += 1
            self._state = self._state_imp
        else:
            self.context.add_to_long_function_name(' ' + token)

    @staticmethod
    def generate_tokens(source_code, _=None):
        return CodeReader.generate_tokens(
            source_code,
            r'|' + r'|'.join(re.escape(s) for s in (
                '..', '->', '<@', '@>', '@lazy', '@fuzzy',
                '@index', '@deterministic')))



''' programming languages of lizard '''

from .java import JavaReader
from .javascript import JavaScriptReader
from .python import PythonReader
from .objc import ObjCReader
from .ttcn import TTCNReader
from .swift import SwiftReader



import unittest
from test.mock import Mock, patch
from lizard_ext.lizardcpre import LizardExtension as CPreprocessor
from .testHelpers import get_cpp_function_list_with_extnesion
from lizard import CodeReader
generate_tokens = CodeReader.generate_tokens


def process_code(code, reader=None):
    reader = reader or Mock(ext=["c"])
    tokens = generate_tokens(code)
    return [t for t in CPreprocessor()(tokens, reader)]

class TestCPreprocessor(unittest.TestCase):

    def test_should_be_after_line_counter(self):
        self.assertEqual(0, CPreprocessor().ordering_index)

    def test_should_repeat_tokens(self):
        tokens = process_code("a b c")
        self.assertEqual(['a', ' ', 'b', ' ', 'c'], tokens)

    def test_should_remove_preprocessor_tokens(self):
        tokens = process_code("#blah a b c\n1")
        self.assertEqual(["\n","1"], tokens)

    def test_should_work_only_for_clike_reader(self):
        reader = Mock(ext=[])
        tokens = process_code("#\n1", reader)
        self.assertEqual(['#',"\n","1"], tokens)

    def test_should_ignore_tokens_in_else(self):
        tokens = process_code("""
        #if x
        1
        #else
        2
        #endif
        3
        """)
        self.assertNotIn("2", tokens)
        self.assertIn("3", tokens)

    def test_should_handle_nested_ifs(self):
        tokens = process_code("""
        #if x
            #if x
            1
            #else
            2
            #endif
            3
        #else
            #if x
            4
            #else
            5
            #endif
            6
        #endif
        """)
        self.assertIn("1", tokens)
        self.assertNotIn("2", tokens)
        self.assertIn("3", tokens)
        self.assertNotIn("4", tokens)
        self.assertNotIn("5", tokens)
        self.assertNotIn("6", tokens)


    def test_should_handle_nested_ifs(self):
        tokens = process_code("""
        #if x
        #else
            #if x
            #endif
        #endif
        1
        """)
        self.assertIn("1", tokens)


def analyze_with_extension(code):
    return get_cpp_function_list_with_extnesion(code, CPreprocessor())


class Test_end_2_end(unittest.TestCase):

    def test_exention_in_user(self):
        result = analyze_with_extension("#if\nxxaa#else\nint a(){}\n#endif")
        self.assertEqual(0, len(result))





import unittest
from .testHelpers import get_cpp_function_list_with_extnesion
from lizard_ext.lizardexitcount import LizardExtension as ExitCounter

class TestFunctionExitCount(unittest.TestCase):

    def test_no_return_should_count_as_1(self):
        result = get_cpp_function_list_with_extnesion("int fun(){}", ExitCounter())
        self.assertEqual(1, result[0].exit_count)

    def test_one_return_should_count_as_1(self):
        result = get_cpp_function_list_with_extnesion("int fun(){return 0;}", ExitCounter())
        self.assertEqual(1, result[0].exit_count)

    def test_two_returns_should_count_as_2(self):
        result = get_cpp_function_list_with_extnesion("int fun(){return 0;return 1;}", ExitCounter())
        self.assertEqual(2, result[0].exit_count)




import unittest
from lizard import CodeReader
generate_tokens = CodeReader.generate_tokens


class Test_generate_tonken(unittest.TestCase):

    def check_tokens(self, source, *expect):
        tokens = generate_tokens(source)
        self.assertEqual(list(expect), tokens)

    def test_empty_string(self):
        self.check_tokens("")

    def test_spaces(self):
        self.check_tokens("\n", "\n")
        self.check_tokens("\n\n", "\n", "\n")
        self.check_tokens(" \n", " ", "\n")

    def test_digits(self):
        self.check_tokens("1", "1")
        self.check_tokens("123", "123")

    def test_operators(self):
        self.check_tokens("-;", '-', ';')
        self.check_tokens("-=", '-=')
        self.check_tokens(">=", '>=')
        self.check_tokens("<=", '<=')
        self.check_tokens("||", '||')

    def test_more(self):
        self.check_tokens("int a{}", 'int', ' ', "a", "{", "}")

    def test_string(self):
        self.check_tokens(r'""', '""')
        self.check_tokens(r'"x\"xx")', '"x\\"xx"', ')')
        self.check_tokens("'\\''", "'\\''")
        self.check_tokens(r"'\\\'", *["'", '\\', '\\', '\\', "'"])

    def test_line_number(self):
        self.check_tokens(r'abc', 'abc')

    def test_line_number2(self):
        tokens = generate_tokens('abc\ndef')
        self.assertTrue('def' in tokens)

    def test_with_mutiple_line_string(self):
        tokens = generate_tokens('"sss\nsss" t')
        self.assertTrue('t' in tokens)


class Test_generate_tonken_for_marcos(unittest.TestCase):

    def test_define(self):
        define =  '''#define xx()\
                       abc'''
        tokens = generate_tokens(define+'''
                    int''')
        self.assertEqual([define, '\n', ' ' * 20, 'int'], tokens)

    def test_include(self):
        tokens = generate_tokens('''#include "abc"''')
        self.assertEqual(['#include "abc"'], tokens)

    def test_if(self):
        tokens = generate_tokens('''#if abc\n''')
        self.assertEqual(['#if abc', '\n'], tokens)

    def test_ifdef(self):
        tokens = generate_tokens('''#ifdef abc\n''')
        self.assertEqual(['#ifdef abc', '\n'], tokens)

    def test_with_line_continuer_define(self):
        tokens = generate_tokens('#define a \\\nb\n t')
        self.assertTrue('t' in tokens)

    def test_define2(self):
        tokens = generate_tokens(r''' # define yyMakeArray(ptr, count, size)     { MakeArray (ptr, count, size); \
                       yyCheckMemory (* ptr); }
                       t
                    ''')
        self.assertTrue('t' in tokens)

    def test_half_comment_following(self):
        comment = '''#define A/*\n*/'''
        tokens = generate_tokens(comment)
        self.assertEqual(2, len(tokens))

    def test_block_comment_in_define(self):
        comment = '''#define A \\\n/*\\\n*/'''
        tokens = generate_tokens(comment)
        self.assertEqual(1, len(tokens))

class Test_generate_tonken_for_comments(unittest.TestCase):

    def test_c_style_comment(self):
        tokens = generate_tokens("/***\n**/")
        self.assertEqual(["/***\n**/"], tokens)

    def test_cpp_style_comment(self):
        tokens = generate_tokens("//aaa\n")
        self.assertEqual(['//aaa', '\n'], tokens)

    def test_cpp_style_comment_with_multiple_lines(self):
        tokens = generate_tokens("//a\\\nb")
        self.assertEqual(['//a\\\nb'], tokens)

    def test_commentedComment(self):
        tokens = generate_tokens(" /*/*/")
        self.assertEqual([' ', "/*/*/"], tokens)

    def test_with_cpp_comments(self):
        tokens = generate_tokens('//abc\n t')
        self.assertTrue('t' in tokens)

    def test_with_c_comments(self):
        tokens = generate_tokens('/*abc\n*/ t')
        self.assertTrue('t' in tokens)

    def test_with_c_comments_with_backslash_in_it(self):
        comment = '/**a/*/'
        tokens = generate_tokens(comment)
        self.assertListEqual([comment], tokens)




import unittest
import inspect
from lizard import analyze_file, FileAnalyzer, get_extensions


def get_ttcn_function_list(source_code):
    return analyze_file.analyze_source_code(
        "a.ttcn", source_code).function_list


class Test_parser_for_TTCN(unittest.TestCase):

    def test_empty(self):
        functions = get_ttcn_function_list("")
        self.assertEqual(0, len(functions))

    def test_no_function(self):
        result = get_ttcn_function_list('''
                module test {
                    const integer i := 1;
                }''')
        self.assertEqual(0, len(result))

    def test_one_function(self):
        result = get_ttcn_function_list('''
                module test_one_func {
                    function fun() {}
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)

    def test_two_function(self):
        result = get_ttcn_function_list('''
                module test_two_func {
                    function fun()
                    {
                    }
                    function fun1()
                    {
                    }
                }
                ''')
        self.assertEqual(2, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun1", result[1].name)
        self.assertEqual(3, result[0].start_line)
        self.assertEqual(5, result[0].end_line)
        self.assertEqual(6, result[1].start_line)
        self.assertEqual(8, result[1].end_line)

    def test_one_testcase(self):
        result = get_ttcn_function_list('''
                module test_one_tc {
                    testcase tc() {}
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("__testcase__tc", result[0].name)

    def test_two_testcase(self):
        result = get_ttcn_function_list('''
                module test_two_tc {
                    testcase tc()
                    {
                    }
                    testcase tc1()
                    {
                    }
                }
                ''')
        self.assertEqual(2, len(result))
        self.assertEqual("__testcase__tc", result[0].name)
        self.assertEqual("__testcase__tc1", result[1].name)
        self.assertEqual(3, result[0].start_line)
        self.assertEqual(5, result[0].end_line)
        self.assertEqual(6, result[1].start_line)
        self.assertEqual(8, result[1].end_line)

    def test_one_function_one_testcase(self):
        result = get_ttcn_function_list('''
                module test_one_func_one_tc {
                    function fun()
                    {
                    }
                    testcase tc()
                    {
                    }
                }
                ''')
        self.assertEqual(2, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("__testcase__tc", result[1].name)
        self.assertEqual(3, result[0].start_line)
        self.assertEqual(5, result[0].end_line)
        self.assertEqual(6, result[1].start_line)
        self.assertEqual(8, result[1].end_line)

    def test_function_with_content(self):
        result = get_ttcn_function_list('''
                module test_func_w_ctx {
                    function fun(inout integer i)
                    {
                        i := call(i);
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( inout integer i )", result[0].long_name)

    def test_testcase_with_content(self):
        result = get_ttcn_function_list('''
                module test_tc_w_ctx {
                    testcase tc(integer i)
                    {
                        i := call(i);
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("__testcase__tc", result[0].name)
        self.assertEqual("__testcase__tc( integer i )", result[0].long_name)

    def test_function_with_runs_on(self):
        result = get_ttcn_function_list('''
                module test_function_with_runs_on {
                    function fun(inout integer i)
                    runs on MTC_CT
                    {
                        i := call(i);
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( inout integer i ) runs on MTC_CT",
                         result[0].long_name)

    def test_function_with_mtc(self):
        result = get_ttcn_function_list('''
                module test_function_with_mtc {
                    function fun( integer i )
                    mtc MTC_CT
                    {
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( integer i ) mtc MTC_CT", result[0].long_name)

    def test_function_with_system(self):
        result = get_ttcn_function_list('''
                module test_function_with_system {
                    function fun( integer i )
                    system MySystemType
                    {
                        var MySystemType v_st := system;
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( integer i ) system MySystemType",
                         result[0].long_name)

    def test_function_with_return(self):
        result = get_ttcn_function_list('''
                module test_function_with_return {
                    function fun( integer i )
                    return integer
                    {
                        return 42;
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( integer i ) return integer",
                         result[0].long_name)

    def test_function_with_return_w_template(self):
        result = get_ttcn_function_list('''
                module test_function_with_return_w_template {
                    function fun( integer i )
                    return template integer
                    {
                        return ?;
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( integer i ) return template integer",
                         result[0].long_name)

    def test_function_in_group(self):
        result = get_ttcn_function_list('''
                module test_function_in_group {
                    group g1 {
                        function fun( integer i )
                        {
                            return 0;
                        }
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( integer i )", result[0].long_name)

    def test_double_slash_within_string(self):
        result = get_ttcn_function_list('''
                module test_double_slash_within_string {
                    function fun()
                    {
                        character a := "\\";
                    }
                }''')
        self.assertEqual(1, len(result))

    def test_function_with_no_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_no_param {
                    function fun()
                    return integer
                    {
                    }
                }''')
        self.assertEqual(0, result[0].parameter_count)

    def test_function_with_1_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_1_param {
                    function fun(boolean b)
                    return integer
                    {
                        return 1;
                    }
                }''')
        self.assertEqual(1, result[0].parameter_count)

    def test_function_with_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_param {
                    function fun(integer i := bb * cc, integer j := cc * dd)
                    return integer
                    {
                        return i * j;
                    }
                }''')
        self.assertEqual(2, result[0].parameter_count)

    def test_function_with_strang_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_strang_param {
                    function fun(boolean b := aa < cc, boolean c := ee < dd)
                    return boolean
                    {
                        return b && c;
                    }
                }''')
        self.assertEqual(2, result[0].parameter_count)

    def test_function_with_template_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_template_param {
                    function fun(out template boolean b)
                    {
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( out template boolean b )", result[0].long_name)

    def test_function_with_lazy_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_lazy_param {
                    function fun(in @lazy integer p_i)
                    {
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( in @lazy integer p_i )", result[0].long_name)

    def test_function_with_fuzzy_param(self):
        result = get_ttcn_function_list('''
                module test_function_with_fuzzy_param {
                    function fun(in @fuzzy integer p_i)
                    {
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("fun( in @fuzzy integer p_i )", result[0].long_name)

    def test_function_deterministic(self):
        result = get_ttcn_function_list('''
                module test_function_deterministic {
                    function @deterministic fun()
                    {
                    }
                }''')
        self.assertEqual(1, len(result))
        self.assertEqual("fun", result[0].name)
        self.assertEqual("@deterministic fun( )", result[0].long_name)

    def test_brakets_before_function(self):
        result = get_ttcn_function_list('''module err{\n()\n}''')
        self.assertEqual(0, len(result))

    def test_underscore(self):
        result = get_ttcn_function_list('''module err{\nfunction _(){}\n}''')
        self.assertEqual(0, len(result))

    def test_digits(self):
        result = get_ttcn_function_list('''module err{\nfunction 42(){}\n}''')
        self.assertEqual(0, len(result))

    def test_template(self):
        result = get_ttcn_function_list('''
                module test_template {
                    template MyMessageType MyTemplate1 (
                        template ( omit ) integer MyFormalParam
                    ):=
                    {
                        field1 := MyFormalParam,
                        field2 := pattern "abc*xyz",
                        field3 := true
                    }
                }''')
        self.assertEqual(0, len(result))


class Test_Preprocessing(unittest.TestCase):

    def test_content_macro_should_be_ignored(self):
        result = get_ttcn_function_list('''
                module test_content_macro_should_be_ignored {
                    #define MTP_CHEC                    \
                       function foo () {                     \
                       }
                } ''')
        self.assertEqual(0, len(result))

    def test_preproc_should_be_ignored_outside_func_impl(self):
        result = get_ttcn_function_list('''
                module test_preproc_should_be_ignored_outside_func_impl {
                    #ifdef MAGIC
                    #endif
                    function foo()
                    {}
                } ''')
        self.assertEqual(1, len(result))

    def test_preprocessor_is_not_function(self):
        result = get_ttcn_function_list('''
                module test_preprocessor_is_not_function {
                    #ifdef A
                    #elif (defined E)
                    #endif
                } ''')
        self.assertEqual(0, len(result))



import unittest
from .testHelpers import get_cpp_function_list_with_extnesion
from lizard_ext.lizardoutside import LizardExtension as CountOutsideComplexity


def analyze_with_outside_extension(code):
    return get_cpp_function_list_with_extnesion(code, CountOutsideComplexity())

class Test_complexity_in_c_marco(unittest.TestCase):

    def test_no_complexity_outside_function_global_cc_should_be_one(self):
        result = analyze_with_outside_extension("")
        self.assertEqual(1, len(result))
        self.assertEqual("*global*", result[0].name)
        self.assertEqual(1, result[0].cyclomatic_complexity)

    def test_complexity_outside_should_be_counted(self):
        result = analyze_with_outside_extension("#if a==b")
        self.assertEqual(2, result[0].cyclomatic_complexity)

    def test_complexity_outside_should_be_counted_when_there_is_function(self):
        result = analyze_with_outside_extension("#if a==b\n void fun() {if(1);}\n #if 1")
        self.assertEqual("*global*", result[1].name)
        self.assertEqual(3, result[1].cyclomatic_complexity)




'''
Language parser for Apple Swift
'''

from lizard import CodeReader, CCppCommentsMixin


class SwiftReader(CodeReader, CCppCommentsMixin):
    # pylint: disable=R0903

    ext = ['swift']
    language_names = ['swift']

    def __init__(self, context):
        super(SwiftReader, self).__init__(context)
        self._state = self._global
        self.br_count = 0

    def _global(self, token):
        if token == 'func':
            self._state = self._function_name
        if token == 'protocol':
            self._state = self._protocol

    def _function_name(self, token):
        self.context.start_new_function(token)
        self._state = self._expect_function_dec

    def _expect_function_dec(self, token):
        if token == '(':
            self._state = self._function_dec

    def _function_dec(self, token):
        if token == ')':
            self._state = self._expect_function_impl
        else:
            self.context.parameter(token)

    def _expect_function_impl(self, token):
        if token == '{':
            self.br_count += 1
            self._state = self._function_impl

    @CodeReader.read_brackets
    def _function_impl(self, _):
        self._state = self._global
        self.context.end_of_function()

    @CodeReader.read_brackets
    def _protocol(self, _):
        self._state = self._global



'''
This is an extension of lizard, that counts the
internal relationship of a class.
'''
from lizard_ext.keywords import IGNORED_WORDS


def delayed_tokens(tokens):
    def extend_(tokens, extendee):
        for token in tokens:
            yield token
        yield extendee

    previous = None
    for token in extend_(tokens, None):
        if previous:
            yield previous, token
        previous = token


def preprocess(tokens):
    saved_token = ""
    for token, next_token in delayed_tokens(tokens):
        if "::" in (token, next_token):
            saved_token += token
        else:
            yield saved_token + token
            saved_token = ""


class FunctionDependenciesCounter(object):
    # pylint: disable=R0903

    def __init__(self, function_builder):
        self.function_builder = function_builder
        self.potential_type = None
        self._state = self._global_state
        self._local_variables = set()
        self.function_processor = None

    def __call__(self, tokens):
        for token in preprocess(tokens):
            if not hasattr(self.function_builder, "dependencies"):
                self.function_builder.dependencies = set()
            self._state(token)
            yield token

    def _global_state(self, token):
        if token[0].isalpha() and token not in IGNORED_WORDS\
             and token not in self._local_variables\
             and token not in self.function_builder.parameters:
            self.potential_type = token
            self._state = self._potential_local_variable_state

    def _potential_local_variable_state(self, token):
        if not token[0].isalpha():
            self.function_builder.dependencies.add(
                self.potential_type)
        else:
            self._local_variables.add(token)
        self._state = self._global_state


class LizardExtension(object):  # pylint: disable=R0903
    # pylint: disable=R0903

    class StopSplitting(BaseException):
        pass

    def __call__(self, tokens, reader):
        try:
            while True:
                if reader.is_in_function():
                    for token in subcall(
                            self.wild(tokens, reader), reader):
                        yield token
                yield next(tokens)
        except self.StopSplitting:
            pass

    def wild(self, tokens, reader):
        try:
            while reader.is_in_function():
                yield next(tokens)
        except StopIteration:
            raise self.StopSplitting


def subcall(tokens, reader):
    fdc = FunctionDependenciesCounter(reader.context.current_function)
    return fdc(tokens)
